
/* [DEBUG] */
#if 0

static void	
set_control_area ( struct vmcb *vmcb )
{
	/* Enable nested paging (See AMD64 manual Vol. 3, p. 488) */
//	vmcb->np_enable = 1; 

	/* Flush TLB on VMRUN (all entries, all ASIDs) */
        vmcb->tlb_control = 1;

	/* To be added in RDTSC and RDTSCP */
	vmcb->tsc_offset = 0; 
	
	/* Guest address space identifier (ASID) */
	vmcb->guest_asid = 1;

	vmcb->general2_intercepts = INTRCPT_VMRUN;

	/* [REF] vol.2, p. 454 */
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 /* 12 Kbytes */ );
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10 /* 8 Kbytes */ );
}

static void
set_descriptors ( struct vmcb *vmcb )
{
	int i;

	/* Generic Segment Descriptor - Legacy Mode 
	 * (AMD64 manual Vol 2, pp. 98-104) */

	/* setup the segment registers and all their hidden states */
	vmcb->cs.sel   = 0xF000; 
	vmcb->cs.attrs.bytes = 0x089b; // 1000 1001 1011
	vmcb->cs.limit = 0xffff;
	vmcb->cs.base  = 0x000F0000;

	// type = 1011 : Data-segment, R(eadable), A(ccessed)
	// S, 
	// P(resent), 
	// G(ranularity)

	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel   = 0x00;
		x->attrs.bytes = 0x0893; // 1000 1001 0011 
		x->limit = 0xffff;
		x->base  = 0x00;

		// type = 0011 : Data-segment, W(ritable), A(ccessed)
		// S, 
		// P(resent), 
		// G(ranularity)
	}
	
	struct seg_selector *dsegs [] = { &vmcb->ldtr, &vmcb->gdtr, &vmcb->tr, NULL };
	for ( i = 0; dsegs [ i ] != NULL; i++ ) {
		struct seg_selector *x = dsegs [ i ];
		x->sel   = 0x00;
		x->attrs.bytes = 0x0000;
		x->limit = 0x0;
		x->base  = 0x00;
	}

	vmcb->idtr.sel   = 0x00;
	vmcb->idtr.attrs.bytes = 0x0000;
	vmcb->idtr.limit = 0x3ff;
	vmcb->idtr.base  = 0x00;
}

static void
set_state_save_area ( struct vmcb *vmcb )
{
	set_descriptors ( vmcb );

	vmcb->cpl = 0; /* must be equal to SS.DPL  */

	vmcb->rax = 0;
	vmcb->rsp = 0;

	vmcb->rflags = 2;

//	vmcb->cr3 = pagetable_get_paddr ( svm_paging_enabled(v) ? v->arch.guest_table : ov->domain->arch.phys_table );
	
	/* (Paged real mode) VMRUN can load a guest value of CR0 with PE = 0 and PG = 1,
	 * a combination that is otherwise illegal (See AMD64 manual vol. 2, p. 445, p. 467). */
//	vmcb->cr0 = X86_CR0_ET | X86_CR0_PG;
	vmcb->cr0 = X86_CR0_ET; 

//	vmcb->cr4 = X86_CR4_PAE;

	vmcb->efer = EFER_SVME;
}

/* Create a new virtual machine with a given OS image.
 *  [input]
 *    - guest_image_start : 
 *    - guest_image_size : 
 *    - guest_pmem_size :
 *  [output]
 *    - vm : 
 */
void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	/* Allocate a new page for storing VMCB.  */
	vm->vmcb = alloc_vmcb ( );

	/* Allocate new pages for physical memory of the guest OS.  */
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );

	/* Copy the OS image to the specified region by interpreting the ELF format.  */
	vm->vmcb->rip = load_elf_image ( guest_image_start, guest_image_size, vm_pmem_start );

	/* Setup multiboot info.  */
	vm->mbi = init_vm_mbi ( vm_pmem_start );

	/* Set Host-level CR3 to use for nested paging.  */
	vm->h_cr3 = create_shadow_table ( vm_pmem_start, vm_pmem_size );
//	vm->vmcb->h_cr3 = vm->h_cr3;

	set_control_area ( vm->vmcb );
	set_state_save_area ( vm->vmcb );

	printf ( "New virtual machine created.\n" ); 
}

